/*****************************************************************
 ** Copyright 1999 Clark Cooper
 ** All rights reserved.
 ** This is free software; you may modify it or redistribute it
 ** under the terms of the perl Artistic license.
 **
 ** The algorithm used here borrows heavily from Larry Wall's
 ** implementation of hashtables in perl.
 */

#include <string.h>
#include <stdlib.h>

#include "hash.h"

#define NEW(type, size) (type*) malloc(size * sizeof(type))

#define MAXBLOCK	1024

/*--------------------------------------------------
  The entry free list uses a mixed vector and linked list
  approach. The following are codes used in the hash field of
  empty entries on the free list to tell by which method to get
  the next free entry. USE_NEXT means that the next field of the
  entry points to the next free entry. USE_AFTER means that the
  next free entry may be found immediately after this entry in memory.
*/
   
#define USE_NEXT	1
#define USE_AFTER	0

#ifndef NULL
#define NULL ((void *) 0)
#endif 

typedef struct _hent	HashEntry;	/* The actual contents of the hash
					   table consists of pointers to
					   this structure. */

struct _hent
{
  unsigned int	hash;		/* The full hash value of the object */
  HashObj	obj;		/* The object being hashed */
  HashEntry	*next;		/* Pointer to next entry that has the
				   same table address as this one */
};

struct _htable
{
  int		entry_count;	/* The number of objects in hash table */
  unsigned int	size;		/* One less than the size of table (2^n - 1) */
  int		kcnt;		/* Number of keys in HashObj */
  KInfo		ki[HASH_MAXKEYS]; /* Per key info */
  HashEntry	**table;	/* Actual hash table */
  HashEntry	*memhead;	/* Pointer to head of list of blocks of entry
				   objects */
  HashEntry	*memtail;	/* Pointer to tail of list of blocks of entry
				   objects */
  HashEntry	*free;		/* Pointer to next free entry object */
};


/*================================================================
** Private functions
*/


/*--------------------------------------------------
  Return an unsigned integer derived from the keys
  of the object.
  */

static unsigned int
hash(HashTable *tab, ConstHashObj obj)
{
  register unsigned int	ret = 0;

  if (tab->kcnt == 0)
    {
      /* HashObj interpreted as char pointer */

      register char	*ptr = (char *) obj;

      while (*ptr)
	ret = ret * 33 + *ptr++;
    }
  else
    {
      int	index;

      for (index = 0; index < tab->kcnt; index++)
	{
	  int	size = tab->ki[index].siz;
	  char		*addr = ((char *) obj) + tab->ki[index].off;

	  /*--------------------------------------------------
	     Size encodes key type + => fixed-width key
	     			   0 => pointer to string
				   - => fixed width but possibly null
				        terminated string
	     See the HASH_KEY... defines in hash.h.
	     */

	  if (size == 0)
	    {
	      register char	*ptr = *((char **) addr);

	      while (*ptr)
		ret = ret * 33 + *ptr++;
	    }
	  else if (size < 0)
	    {
	      register char	*ptr = addr;

	      size = -size;
	      for (; *ptr && size > 0; size--)
		ret = ret * 33 + *ptr++;
	    }
	  else
	    {
	      register char	*ptr = addr;

	      for (; size > 0; size--)
		ret = ret * 33 + *ptr++;
	    }
	}
    }

  return ret;
}  /* End hash */

/*--------------------------------------------------
  Test for equality of two objects. Returns 1 if
  objects are equal, otherwise returns 0
  */

static int
hasheq(HashTable *ht, ConstHashObj a, ConstHashObj b)
{
  int	index;

  if (ht->kcnt == 0)
    {
      register const char	*ap = a;
      register const char	*bp = b;

      /* Pointer to string case, compare each byte
	 for equality until terminating null byte reached */

      for (; *ap == *bp; ap++, bp++)
	{
	  if (! *ap)
	    return 1;
	}
      return 0;
    }

  for (index = 0; index < ht->kcnt; index++)
    {
      int	siz = ht->ki[index].siz;
      char		*addr_a = ((char *) a) + ht->ki[index].off;
      char		*addr_b = ((char *) b) + ht->ki[index].off;

      if (siz == 0)
	{
	  register char	*ap = *((char **) addr_a);
	  register char	*bp = *((char **) addr_b);

	  /* Pointer to string case, compare each byte
	     for equality until terminating null byte reached */

	  for (; *ap == *bp; ap++, bp++)
	    {
	      if (! *ap)
		goto nextkey;
	    }
	  return 0;
	}
      else if (siz < 0)
	{
	  register int	index;
	  register char *ap = addr_a;
	  register char *bp = addr_b;

	  /* Fixed width but possibly null terminated case */

	  int		limit = -siz;

	  for (index = 0; index < limit; ap++, bp++, index++)
	    {
	      if (*ap != *bp)
		return 0;

	      if (*ap == '\0')
		break;
	    }
	}
      else
	{
	  /* Simple fixed width; each byte must match */

	  if (memcmp(addr_a, addr_b, siz))
	    return 0;
	}

    nextkey:
      ;
    }

  return 1;
}  /* End hasheq */

/*--------------------------------------------------
  Allocate more hashentries. If entry_count > size,
  double the amount of room available to the hash table.
  Return 0 if successful, 1 if we couldn't allocate space
  for the expansion.
  */

static int
hashgrow(HashTable *ht)
{
  int		blocksize;
  HashEntry	*newblock;

  /* Note that in this implementation, the number of entries can be
     greater than the table size. We increase the size as we go along
     to improve performance. Too small a table size turns a hash
     search into a linear search, since for a small table we usually
     have to travel down a linked list sequentially to find the
     object we're looking for.

     The point at which to increase the table size is really a tuning
     parameter.
     */

  if (ht->entry_count > ht->size)
    {
      unsigned int	oldsize = ht->size + 1;
      unsigned int	newsize = oldsize << 1;
      unsigned int	newmask = newsize - 1;
      HashEntry		**newtab = NEW(HashEntry *, newsize);
      HashEntry		**slot;
      HashEntry		**limit;

      if (! newtab)
	return 1;

      /* Copy old table to base of new location */

      memcpy(newtab, ht->table, sizeof(HashEntry *) * oldsize);

      limit = &newtab[oldsize];

      free (ht->table);

      /* By doubling the size of the table, an additional bit of
	 the hash value must be taken into account to derive the
	 index at which to store the object. So we have to visit
	 existing entries, looking for entries that really belong
	 in the new upper half

	 Suppose an object has a hash value of 25. With a table size
	 of 8, it belongs at slot 1, but with a table size of 16, it
	 belongs at slot 9. After expansion, it has to be moved or it
	 won't be found.

	 An object that has a zero in the next significant bit is already
	 in the right location. So, using our previous example, an
	 object with a hash value of 17 is already in its proper slot of
	 1 after the memcpy above.

	 Since the oldsize is a power of two, it is a mask for checking
	 the state of the next bit.
	 */

      for (slot = newtab; slot < limit; slot++)
	{
	  HashEntry		**newstep = slot + oldsize;
	  HashEntry		**step = slot;
	  HashEntry		*entry;

	  while ((entry = *step) != NULL)
	    {
	      if (entry->hash & oldsize)
		{
		  /* Entry belongs in upper half. Remove it from this
		     list and put it on a list in the upper half */

		  *newstep = entry;
		  *step = entry->next;
		  newstep = &entry->next;
		}
	      else
		step = &entry->next;
	    }

	  /* This either terminates the list or intializes the slot
	     in the new half to be empty.
	     */

	  *newstep = NULL;
	}

      ht->table = newtab;
      ht->size = newmask;
    }

  /* For small sizes, give them what they want + the additional entries
     needed to keep track of the blocks.
     */

  blocksize = ht->size > MAXBLOCK ? MAXBLOCK : (ht->size + 2);

  newblock = NEW(HashEntry, blocksize);
  if (! newblock)
    return 1;

  /* This intializes the entries to empty and to using the vector
     (USE_AFTER) method of getting the next free entry */

  memset(newblock, 0, blocksize * sizeof(HashEntry));

  /* This function is only called when ht->free is NULL, so we don't
     need to save it */

  ht->free = newblock;

  /* The last entry in each block is used to manage the blocks themselves.
     The next to last entry (which is the last entry in the block that can
     be used by the table) has to use the linked list (USE_NEXT) method
     and has to point to an empty list. Therefore, after this entry has
     been used, the free pointer will be null, triggering the next call
     to hashgrow.
     */

  newblock += blocksize - 2;
  newblock->hash = USE_NEXT;
  newblock++;

  newblock->obj = (HashObj) ht->free;
  newblock->hash = blocksize;

  if (ht->memtail)
    {
      ht->memtail->next = newblock;
      ht->memtail = newblock;
    }
  else
    ht->memhead = ht->memtail = newblock;

  return 0;
}  /* End hashgrow */

/*================================================================
** Public functions
*/

/*--------------------------------------------------
  Return a hash table with room for at least initial_size
  elements. If kcnt <= 0, kin may be NULL and the key is either a
  null-terminated string or a structure that begins with a null-terminated
  string (e.g. struct {char x[16]; int foo} could have kcnt == 0 if all
  instances of x were null terminated, struct {char *y; int foo;} couldn't
  have kcnt == 0).

  Otherwise kcnt is the number of keys specified in kin. Except in the
  simple case where kcnt == 0, newhash shouldn't be called directly.
  Instead callers should use the macros HASH_BEGIN, HASH_KEY, HASH_KEY_STR,
  and HASH_END defined in hash.h.

  Returns NULL if there's not enough memory to allocate some part of
  the hashtable or if kcnt > HASH_MAXKEYS.
  */

HashTable *
newhash(int initial_size, int kcnt, KInfo *kinf)
{
  HashTable	*ret = NEW(HashTable, 1);
  unsigned int	allocsiz;

  /* Calculate power of 2 larger than initial_size */

  if (ret == NULL || kcnt > HASH_MAXKEYS)
    return NULL;

  if (initial_size <= 0)
    {
      allocsiz = 16;
    }
  else
    {
      int		bits;
      unsigned int	chsiz = (unsigned int) initial_size;

      for (bits = 0; chsiz; bits++, chsiz >>= 1)
	;

      if (! bits)
	return NULL;

      allocsiz = 1<<(bits);
    }

  ret->table = NEW(HashEntry *, allocsiz);
  if (ret->table == NULL)
    {
      free(ret);
      return NULL;
    }

  /* With free being set to NULL, the first insert will cause a call of
     hashgrow to actually allocate some entries.
     */

  ret->memhead = ret->memtail = NULL;
  ret->free = NULL;
  memset(ret->table, 0, allocsiz * sizeof(HashEntry *));

  ret->entry_count = 0;
  ret->size = allocsiz - 1;
  if (kcnt <= 0)
    ret->kcnt = 0;
  else
    {
      int	i;

      ret->kcnt = kcnt;

      for (i = 0; i < kcnt; i++)
	{
	  ret->ki[i].off = kinf[i].off;
	  ret->ki[i].siz = kinf[i].siz;
	}
    }

  return ret;
}  /* End newhash */


/*--------------------------------------------------
  Return the HashObj matching the key. If no entry matches,
  return NULL.
  */

HashObj
hashfind(HashTable *ht, ConstHashObj key)
{
  unsigned int	hsh = hash(ht, key);
  HashEntry	*step = ht->table[hsh & ht->size];

  for (; step; step = step->next)
    {
      if (step->hash == hsh && hasheq(ht, key, step->obj))
	return step->obj;
    }

  return NULL;
}  /* End hashfind */

/*--------------------------------------------------
  If a HashObj is found matching the key, return that entry.
  Otherwise insert the entry provided as a key and return it.
  */

HashObj
hashinsert(HashTable *ht, ConstHashObj key)
{
  unsigned int	hsh = hash(ht, key);
  HashEntry	**slot = &ht->table[hsh & ht->size];
  HashEntry	**step;
  HashEntry	*entry;

  for (step = slot; (entry = *step) != NULL; step = &entry->next)
    {
      if (entry->hash == hsh && hasheq(ht, key, entry->obj))
	return entry->obj;
    }

  /* No match for key, need to insert it */

  if (! ht->free)
    {
      int	orig_size = ht->size;

      if (hashgrow(ht))
	return NULL;

      /* If the table size changed, slot is no longer
	 a reliable place to insert this entry */

      if (ht->size != orig_size)
	{
	  slot = &ht->table[hsh & ht->size];
	}
    }

  entry = ht->free;

  if (entry->hash == USE_NEXT)
    ht->free = entry->next;
  else
    ht->free = entry + 1;

  entry->hash = hsh;
  entry->obj = (HashObj) key;	/* OK to cast this to a non const since we
				   never change the contents of what user
				   passed us. Since the obj field is used
				   for block maintenance, we don't want
				   it to be a const */

  /* New entry is put on the front of the list */

  entry->next = *slot;
  *slot = entry;

  ht->entry_count++;

  return entry->obj;
}  /* End hashinsert */


/*--------------------------------------------------
  If an entry is found matching the key, delete that entry from
  the table and return it. Otherwise return NULL.
  */

HashObj
hashdelete(HashTable *ht, ConstHashObj key)
{
  unsigned int	hsh = hash(ht, key);
  HashEntry	**step = &ht->table[hsh & ht->size];
  HashEntry	*entry;

  for (; (entry = *step) != NULL; step = &entry->next)
    {
      if (entry->hash == hsh && hasheq(ht, key, entry->obj))
	{
	  HashObj	ret = entry->obj;

	  /* Remove entry from list, empty it, and place it on the
	     free list */

	  *step = entry->next;
	  entry->next = ht->free;
	  entry->hash = USE_NEXT;
	  entry->obj = NULL;
	  ht->free = entry;
	  ht->entry_count--;

	  return ret;
	}
    }

  return NULL;
}  /* End hashdelete */

/*--------------------------------------------------
  Delete all entries from hash table.
  */

void
hashclear(HashTable *ht)
{
  HashEntry	*step;
  HashEntry	**fill = &ht->free;

  /* The last entry in each block contains a pointer to the
     beginning of the block in the obj field and the block size
     (in units of entries) in the hash field.
     */

  for (step = ht->memhead; step; step = step->next)
    {
      /* Clear the block (empty entries and have them USE_AFTER)
         We use the size - 1, since the last entry actually contains
	 block management information. */

      memset(step->obj, 0, sizeof(HashEntry) * (step->hash - 1));
      
      /* Next to last entry (last usable entry) uses the USE_NEXT
	 method and should chain together the blocks */

      step[-1].hash = USE_NEXT;
      *fill = (HashEntry *) step->obj;
      fill = &step[-1].next;
    }

  *fill = NULL;

  ht->entry_count = 0;
  memset(ht->table, 0, sizeof(HashEntry *) * (ht->size + 1));
}  /* End hashclear */

/*--------------------------------------------------
  Free hash table and internal memory assocated with it.
  This does not free the memory of HashObj entries, which
  are the responsibility of the caller. If the hash table
  contains the only references to the object, you may free
  the objects (before calling hashclear or hashfree) by
  iterating over them with the hashnext function described below.

  */

void
hashfree(HashTable *ht)
{
  HashEntry	*step = ht->memhead;

  while (step)
    {
      HashEntry	*next = step->next;

      free(step->obj);
      step = next;
    }
    
  free(ht->table);
  free(ht);
}  /* End hashfree */

/*--------------------------------------------------
  Return the number of entries in the table.
  */

extern int
hashcount(HashTable *ht)
{
  return ht->entry_count;
}  /* End hashcount */


/*--------------------------------------------------
  Place a pointer to a newly allocated vector of hash objects
  containing all the entries in the table in the variable pointed
  at by vec. Return value is the number of objects in this vector.
  The caller is responsible for freeing the vector.

  If there isn't enough memory to allocate the vector, then a -1
  is returned and vec is set to NULL. If the table is empty, then
  the variable pointed at by vec isn't changed.
  */

extern int
hashvector(HashTable *ht, HashObj **vec)
{
  HashObj *fill;
  HashEntry *mb;
  HashEntry *step;

  if (ht->entry_count == 0)
    return 0;

  fill = *vec = (HashObj *) NEW(HashObj, ht->entry_count);
  if (*vec == NULL)
    return -1;

  mb = ht->memhead;
  step = mb->obj;

  
  for (mb = ht->memhead; mb; mb = mb->next) {
    for (step = mb->obj; step < mb; step++) {
      if (step->obj) {
	*fill++ = step->obj;
      }
      else if (step->hash == USE_AFTER) {
	goto done;
      }
    }
  }

 done:
  return ht->entry_count;
}  /* End hashvector */

/*--------------------------------------------------
  This function is used for iterating over all the
  entries in the hash table. To start off, set the
  value pointed to by ITER to NULL. The function will
  return the next hash object from the table. It returns
  NULL when there are no more entries (and sets the HashIterator
  to which ITER points to NULL). Note that hash tables
  do not order their objects. Note that ITER should not be
  NULL. (However hashnext will return NULL if you
  do pass a NULL value for ITER).
  */

HashObj
hashnext(HashTable *ht, HashIterator *iter)
{
  HashEntry	*mb;
  HashEntry	*step;

  if (! iter)
    return NULL;

  mb = ht->memhead;

  if (! mb)
    {
      *iter = NULL;
      return NULL;
    }

  step = (HashEntry *) *iter;

  if (step)
    {
      /* If step is non-NULL, then we're in the middle of an
	 iteration throught the elements of the table */

      /* Find the memory block that contains the current
	 entry */

      for (; mb; mb = mb->next)
	{
	  if (step >= (HashEntry *) mb->obj && step <= mb)
	    break;
	}

      /* If the current entry points to the block management
	 entry at the end of the block, move it along to
	 the next block */

      if (step == mb)
	{
	  mb = mb->next;
	  if (mb)
	    step = (HashEntry *) mb->obj;
	}

      if (! mb)
	{
	  /* No more blocks => end of iteration */

	  *iter = NULL;
	  return NULL;
	}
    }
  else
    step = (HashEntry *) mb->obj;

  for (;;)
    {
      /* Find the first non-empty entry and return it */

      for (; step < mb; step++)
	{
	  if (step->obj)
	    {
	      /* Next iteration will look at entry after this one
		 in memory block */

	      *iter = (HashIterator) (step + 1);
	      return step->obj;
	    }
	  else if (step->hash == USE_AFTER)
	    {
	      /* An empty USE_AFTER entry implies that the rest of
		 this memory block is virgin and therfore that there
		 are no more entries.
		 */

	      *iter = NULL;
	      return NULL;
	    }
	}

      /* When there are no more memory blocks, we're through */

      if (! mb->next)
	break;

      mb = mb->next;
      step = (HashEntry *) mb->obj;
    }
  
  *iter = NULL;
  return NULL;
}  /* End hashnext */


/*================================================================
  The following are for debugging support.
*/


/*--------------------------------------------------
  Print out statistics on hashtable.
  */
void
hashstats(HashTable *ht)
{
  HashEntry	**slot = ht->table;
  HashEntry	**limit = &ht->table[ht->size + 1];
  double	runsqsum = 0.0;
  double	mean;
  int		runsum = 0;
  int		maxrun = 0;
  int		norun = 0;

  printf("Size %d  Entries %d Keys %d\n", ht->size, ht->entry_count,
	 ht->kcnt);

  for (; slot < limit; slot++)
    {
      HashEntry	*step = *slot;
      int	run;

      if (! step)
	continue;

      norun++;

      for (run = -1; step; step = step->next)
	run++;

      runsum += run;
      runsqsum += run * run;

      if (run > maxrun)
	maxrun = run;
    }

  mean = runsum;
  mean /= ht->entry_count;

  printf("Mean run = %6.2f  Var. = %6.2f  Max = %d Zero-run entries = %d\n",
	 mean, runsqsum/ht->entry_count - mean*mean,
	 maxrun, norun);

}  /* End hashstats */
